"""
Reads in an HTCondor log file and parses the relevant data from it, splitting it into events and formatting those
into messages that are sent to the workspaces.workflow-service.workflow-status RabbitMQ exchange. Also input is
a workflow request ID that is sent alongside the event data.
"""

import argparse
import functools
import logging
import os
import re
import signal
import sys
import time
import tracemalloc
from pathlib import Path
from typing import Any, Callable, Dict, List, TextIO, Tuple, Union

import pendulum
from channels.amqp_helpers import workflow_events

from workspaces.schema import WorkflowEvent, WorkflowEventType

from ._version import ___version___ as VERSION

logger = logging.getLogger("wf_monitor")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def log_decorator_factory(msg: str):
    """
    Factory that produces log decorators with a given message

    :param msg: Message to be logged
    :return: Log decorator
    """

    def log_decorator(f: Callable) -> Callable:
        """
        Decorator that logs a message before and after a function is called

        :param f: Function to be decorated
        :return: Wrapper function
        """

        @functools.wraps(f)
        def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
            """
            Decorator boilerplate: wrapper function for log decorator

            :param args: args for function f
            :param kwargs: kwargs for function f
            :return: Return value of the function
            """
            logger.info(msg)
            result = f(*args, **kwargs)
            logger.info("Done.")
            return result

        return wrapper

    return log_decorator


def timeout_handler(signum: int, frame: tracemalloc.Frame) -> None:
    """
    Signal handler function for handling a SIGALRM that arrives when WorkflowMonitor.read_log()
    times out.

    :param signum: Signal number (would be signal.SIGALRM in this case)
    :param frame: Stack frame the signal was sent from
    """
    msg = (
        "No HTCondor log file found in timeout period. The job either did "
        "not log any output or the path to the log was not entered correctly."
    )
    raise TimeoutError(msg)


# Timeout handling solution found here:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
signal.signal(signal.SIGALRM, timeout_handler)


def get_job_name(body: str) -> str:
    """
    Parses a log entry body string for the name of the job currently executing

    :param body: Body of the log entry
    :return: Job name
    """
    r_body_parse = r"[a-zA-Z0-9-_ ]+: (?P<jobname>[a-zA-Z0-9-_]+)"
    body_match = re.search(r_body_parse, body)
    return body_match.group("jobname")


def get_year(logfile_path: Path) -> str:
    """
    Parses year from log file creation date

    :param logfile_path: Path to log file
    :return: Year log was created as a string
    """
    status = os.stat(logfile_path)
    date = pendulum.from_timestamp(status.st_ctime)
    return str(date.year)


def fmt_timestamp(timestamp: str, year: str) -> str:
    """
    Format timestamp string into ISO 8601 format
    2020-08-27T21:45:09+07:00

    :param timestamp: Unformatted timestamp string
    :param year: Year HTCondor log was created as a string
    :return: Formatted timestamp string
    """
    r_nums = r"[0-9]{2}"
    r_timestamp = (
        rf"(?P<mon>{r_nums})/(?P<d>{r_nums}) "
        rf"(?P<h>{r_nums}):(?P<m>{r_nums}):(?P<s>{r_nums})"
    )
    match = re.search(r_timestamp, timestamp)

    return "{year}-{mon}-{day}T{h}:{m}:{s}+07:00".format(
        year=year,
        mon=match.group("mon"),
        day=match.group("d"),
        h=match.group("h"),
        m=match.group("m"),
        s=match.group("s"),
    )


def get_retval(body: str) -> Union[int, None]:
    """
    Parse return value from job terminated event

    :param body: Body of event log entry
    :return: Return value as an int
    """
    r_retval = r"return value (?P<retval>[0-9]+)"

    retval = re.search(r_retval, body)
    if retval:
        return int(retval.group("retval"))
    else:
        return


class WorkflowMonitor:
    """
    Class that monitors the events in a given workflow via the log file produced by it
    """

    def __init__(self, logfile_path: str, workflow_request_id: int, timeout: int = 60):
        self.logfile_path = Path(logfile_path)
        self.workflow_request_id = workflow_request_id
        # TODO: If dag, find out number of jobs
        self.num_jobs = 1
        self.jobs = {}
        # Leaving the list of events in for now; used for testing; might remove later
        self.events = []
        # Leaving the log in for now; used for testing; might remove later
        self.log = ""

        self.read_htcondor_log(timeout)

        logger.info(f"Finished monitoring {self.logfile_path}.")

    @log_decorator_factory("Looking for HTCondor log file...")
    def log_exists(self, timeout: int) -> bool:
        """
        Checks file location repeatedly and returns true if log file is found.

        :param timeout: Amount of time in seconds to wait for the log to appear.
        :raises TimeoutError: If log file not found within timeout seconds
        :return: True if log file found; else raises TimeoutError
        # TODO: Make sure timeout is working as intended
        """
        signal.alarm(timeout)
        while not os.path.exists(self.logfile_path):
            logger.info("Attempting to read log file...")
            time.sleep(3)
        return True

    @log_decorator_factory("Reading HTCondor log...")
    def read_htcondor_log(self, timeout: int):
        """
        Reads HTCondor log as it's being written. When a full event is read in,
        the event is parsed and sent to the workflow events AMQP exchange

        :param timeout: Time in seconds to wait for the job to write to the log
        """
        exists = False
        try:
            exists = self.log_exists(timeout)
        except TimeoutError as e:
            # log_exists() timed out
            logger.error(e)
            quit(-1)

        if exists:
            logger.info("Log file found.")
            event_str = ""
            with open(self.logfile_path, "r") as f:
                # Read while file is being written to
                while True:
                    if self.num_jobs == 0:
                        return

                    where = f.tell()
                    line = f.readline()
                    if not line:
                        time.sleep(1)
                        f.seek(where)
                    else:
                        if line == "...\n":
                            # Line is at the end of an event
                            event_str += line
                            # Parse and send event
                            try:
                                event = self.parse_event(event_str)
                            except ValueError as e:
                                # Event not well-formed
                                logger.error(e)
                                continue
                            send_event(event)
                            # Empty contents to prepare for next event
                            self.log += event_str
                            event_str = ""
                        elif line == "\n":
                            # Line empty
                            time.sleep(1)
                        else:
                            # Line is anywhere else
                            event_str += line

    @log_decorator_factory("Parsing HTCondor log...")
    def parse_event(self, event_str: str) -> WorkflowEvent:
        """
        Parse event for relevant details:

        - HTCondor event number
        - HTCondor job number
        - Timestamp of event
        - Description of event
        - Body with additional information

        :return: List of WorkflowEvent objects
        """
        # Regex that matches an event entry in the HTCondor log file.
        r_htc_event = (
            r"(?P<eventnum>[0-9]{3}) \((?P<jobnum>[0-9]{4})\.[0-9]{3}\.[0-9]{3}\) "
            r"(?P<timestamp>[0-9]{2}/[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}) "
            r"(?P<desc>[a-zA-Z0-9 ?&:=<>\-._]+)\n"
            r"(?P<body>[\t ]+[^\.]*\n)*"
            r"..."
        )
        match = re.match(r_htc_event, event_str)
        if match:
            # Event is well-formed
            job_id = match.group("jobnum")
            if match.group("eventnum") == "000":
                # Populate jobs dict with job name
                self.jobs[job_id] = get_job_name(match.group("body"))

            try:
                event_type = WorkflowEventType(int(match.group("eventnum")))
            except ValueError:
                event_type = WorkflowEventType.OTHER

            if event_type == WorkflowEventType.TERMINATED:
                retval = get_retval(match.group("body"))
            else:
                retval = None

            year = get_year(self.logfile_path)
            event = WorkflowEvent(
                workflow_request_id=self.workflow_request_id,
                job_name=self.jobs[job_id],
                condor_job_id=job_id,
                event_type=event_type,
                timestamp=fmt_timestamp(match.group("timestamp"), year),
                log=match.group(0),
                retval=retval,
            )
        else:
            # Regex didn't return any matches (event not well-formed)
            raise ValueError(
                f"HTCondor event not well-formatted. Event contents:\n{event_str}"
            )

        if event.type == WorkflowEventType.TERMINATED:
            self.num_jobs -= 1

        self.events.append(event)
        return event

    def __str__(self):
        return f"WorkflowMonitor, monitoring {self.logfile_path} that has events {self.events}"

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.logfile_path}, {self.events})"


_DESCRIPTION = (
    "Workspaces workflow monitor, version {}. Monitor execution of a workflow from "
    "the command line."
)


def send_event(event: WorkflowEvent):
    """
    Send workflow event to AMQP exchange

    :param event: Workflow event
    """
    decorated_send = log_decorator_factory("Sending event...")(workflow_events.send)
    decorated_send(event)


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for command-line arguments

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description=_DESCRIPTION.format(VERSION),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-P",
        "--profile",
        action="store",
        required=False,
        default="nmtest",
        help="profile name to use; e.g. test, production; DEFAULT: nmtest",
    )
    parser.add_argument(
        "log_path", action="store", help="path to the workflow's HTCondor log"
    )
    parser.add_argument(
        "workflow_request_id",
        action="store",
        help="ID of the workflow request that ran the workflow",
    )

    return parser


def main():
    # Parse command-line args
    args = make_arg_parser().parse_args()
    WorkflowMonitor(args.log_path, int(args.workflow_request_id))
